/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-tlsdane-0.c,v 1.9 2014/03/04 05:38:33 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/test.h"
#include "sm/io.h"
#include "sm/ctype.h"
#include "sm/tls.h"
#include "sm/tlsdane.h"
#include "sm/tlsfp.h"

#if MTA_USE_TLS && MTA_TLS_DANE

static sm_ret_T
testdane(const char *certfile)
{
	sm_ret_T ret;
	dane_ctx_P dane_ctx;
	tls_ctx_P tls_ctx;
	X509 *cert;
	sm_str_P fp;
	FILE *cf;

	cert = NULL;
	cf = NULL;
	ret = sm_tls_init_library(&tls_ctx, false);
	SM_TEST_E(sm_is_success(ret));

	ret = dane_init(&dane_ctx);
	SM_TEST_E(sm_is_success(ret));

	fp = sm_str_new(NULL, 512, 1024);
	SM_TEST_E(NULL != fp);

	tls_ctx->tls_ssl_ctx = SSL_CTX_new(SSLv23_server_method());
	SM_TEST_E(NULL != tls_ctx->tls_ssl_ctx);

	cf = fopen(certfile, "r");
	SM_TEST_E(NULL != cf);
	cert = PEM_read_X509(cf, NULL, NULL, NULL);
	SM_TEST_E(NULL != cert);

	sm_str_clr(fp);
	ret = tls_pkey_md(cert, "sha256", fp);
	SM_TEST_E(sm_is_success(ret));

	sm_io_fprintf(smioout, "fp=%0.*hhy\n", (int) sm_str_getlen(fp), sm_str_data(fp));
	sm_io_flush(smioout);

	/* other clean up? cert, tls_ctx, ... */
  error:
	if (cf != NULL) {
		fclose(cf);
		cf = NULL;
	}
	return ret;
}

#endif /* MTA_USE_TLS && MTA_TLS_DANE */

static void
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "%s: usage:\n", prg);
}

int
main(int argc, char *argv[])
{
	int c;

	while ((c = getopt(argc, argv, "")) != -1) {
		switch (c) {
		  default:
			usage(argv[0]);
			return(1);
		}
	}
	sm_test_begin(argc, argv, "test tlsdane 0");
#if MTA_USE_TLS && MTA_TLS_DANE
	if (argc > 1)	/* pass entire array? */
		testdane(argv[1]);
#endif
	return sm_test_end();
}
