#!/bin/sh
# $Id: t-ehlo-ipv6-0.sh,v 1.3 2014/02/26 13:59:23 ca Exp $
#
# ----------------------------------------
# test EHLO checks
# uses smtpcnf.sh.
# ----------------------------------------
#

SD=`dirname $0`
if test "${SD}" = "."
then
   SD=../chkmts
fi
. ${SD}/common.sh

if [ X"${MTA_NO_IPV6_TEST}" != "X" ]
then
  echo "SKIPPED: $0: MTA_NO_IPV6_TEST set"
  exit 0
fi

chkcto smtps MTA_NETINET6

if test X"`uname`" != "XOSF1"
then
while getopts s:V FLAG
do
  case "${FLAG}" in
    s) SKIP="${SKIP} ${OPTARG}";;
    V) VERBOSE=true;;
  esac
done
shift `expr ${OPTIND} - 1`
fi

# cd to right directory (where the programs will be executed).
cd ../qmgr || exit 1

${CAT} > ${SMXCNF} <<EOF
qmgr { Log_Level = 12; wait_for_server=4; wait_for_client=4; }
smar { Log_Level = 15; nameserver = ${NS}; }
smtpc { Log_Level = 12; remote_port=${SNKPORT}; wait_for_server=4; }
smtps { Log_Level = 12;
  flags = { strict_ehlo_checks, require_EHLO_before_MAIL };
daemon_address = localhost:${SRVPORT};
wait_for_server=4;
}
EOF

# create aliases
${CAT} > ${ALI} <<EOF
abuse:	local:
postmaster:	local:
user:	local:
EOF
# create map
rm -f ${ALIMAP}
${MM} < ${ALI}

S=sender@local.host
R=rcpt@local.host  

# ----------------------------------------
# allow IPv6: EHLO
TEST=1
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
if ${SHELL} ${SD}/smtpcnf.sh -Z 'smtpc2' -c "-e '[IPv6:::1]'" -S$S -R$R -E 1 -X 1 -A 1
then
  :
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
else
   ${VERBOSE} && echo "PASS: ${TEST}"
fi
fi

# ----------------------------------------
# allow IPv6: EHLO
TEST=2
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
if ${SHELL} ${SD}/smtpcnf.sh -Z 'smtpc2' -c "-e '[IPv6:2001:470:6f:50a::3]'" -S$S -R$R -E 1 -X 1 -A 1
then
  :
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
else
   ${VERBOSE} && echo "PASS: ${TEST}"
fi
fi

# ----------------------------------------
# broken IPv6: EHLO
TEST=3
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
if ${SHELL} ${SD}/smtpcnf.sh -Z 'smtpc2' -c "-e '[IPv6:[2001:470:6f:50a::3]'" -S$S -R$R -E 0 -X 0 -A 0
then
  :
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
else
   ${VERBOSE} && echo "PASS: ${TEST}"
fi
fi

# ----------------------------------------
#rm -f ${ACCMAP} ${ALIMAP} ${MTMAP}

# ----------------------------------------
# end
if test ${ERRS} -eq 0
then
  exit 0
else
  echo "$0: ${ERRS} error(s)"
  exit 1
fi
