/*
 * Copyright (c) 2002, 2003, 2005 Sendmail, Inc. and its suppliers.
 *      All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 *	$Id: adb.h,v 1.22 2013/11/07 05:26:46 ca Exp $
 */

#ifndef AQ_ADB_H
#define AQ_ADB_H 1

#include "sm/generic.h"
#include "sm/pthread.h"
#include "sm/mta.h"
#include "sm/actdb.h"

#if SM_THT_PERF
# define aq_thtincradd(aq_ctx)	((aq_ctx)->aq_adds++)
# define aq_thtincrlu(aq_ctx)	((aq_ctx)->aq_lookups++)
# define aq_thtincrdel(aq_ctx)	((aq_ctx)->aq_deletes++)
# define THTINCRACOL(aq_ctx, hd)	do { \
		if (!LIST_EMPTY(hd))	\
			((aq_ctx)->aq_acollisions++);	\
	} while (0)
# define aq_thtincrlcol(aq_ctx)	((aq_ctx)->aq_lcollisions++)
# define THTINCRDCOL(aq_ctx, aq_node)	do {	\
	if (LIST_NEXT(aq_node, ecn_htl) != LIST_END(aq_ctx) || \
		(aq_node) != LIST_FIRST(aq_ctx))	\
			(aq_ctx)->aq_dcollisions++;			\
	} while (0)
# define aq_thtincrgrows(aq_ctx)	((aq_ctx)->aq_grows++)
# define aq_thtinitchain(aq_ctx)	(aq_ctx)->aq_chainlen = 0
# define THTINCRCHAIN(aq_ctx)	do {	\
		if (++(aq_ctx)->aq_chainlen > (aq_ctx)->aq_chainmax)	\
			(aq_ctx)->aq_chainmax = (aq_ctx)->aq_chainlen;\
	} while (0)
# define THTCHAINLEN(aq_ctx, hd, aq_node) do {	\
		aq_node_P h;		\
		(aq_ctx)->aq_chainlen = 0;	\
		h = aq_node;			\
		while (h != LIST_END(hd)) {				\
			++(aq_ctx)->aq_chainlen;	\
			h = LIST_NEXT(h, ecn_htl);	\
		}				\
		if ((aq_ctx)->aq_chainlen > (aq_ctx)->aq_chainmax)	\
			(aq_ctx)->aq_chainmax = (aq_ctx)->aq_chainlen;	\
	} while (0)

#else /* SM_HT_PERF */
# define aq_thtincradd(aq_ctx)
# define aq_thtincrlu(aq_ctx)
# define aq_thtincrdel(aq_ctx)
# define THTINCRACOL(aq_ctx, hd)
# define aq_thtincrlcol(aq_ctx)
# define THTINCRDCOL(aq_ctx, aq_node)
# define aq_thtincrgrows(aq_ctx)
# define aq_thtinitchain(aq_ctx)
# define THTINCRCHAIN(aq_ctx)
# define THTCHAINLEN(aq_ctx, hd, aq_node)
#endif /* SM_HT_PERF */


#if SM_RHT_PERF
# define aq_rhtincradd(aq_ctx)	((aq_ctx)->aq_adds++)
# define aq_rhtincrlu(aq_ctx)	((aq_ctx)->aq_lookups++)
# define aq_rhtincrdel(aq_ctx)	((aq_ctx)->aq_deletes++)
# define RHTINCRACOL(aq_ctx, hd)	do { \
		if (!LIST_EMPTY(hd))	\
			((aq_ctx)->aq_acollisions++);	\
	} while (0)
# define aq_rhtincrlcol(aq_ctx)	((aq_ctx)->aq_lcollisions++)
# define RHTINCRDCOL(aq_ctx, aq_node)	do {	\
	if (LIST_NEXT(aq_node, ecn_htl) != LIST_END(aq_ctx) || \
		(aq_node) != LIST_FIRST(aq_ctx))	\
			(aq_ctx)->aq_dcollisions++;			\
	} while (0)
# define aq_rhtincrgrows(aq_ctx)	((aq_ctx)->aq_grows++)
# define aq_rhtinitchain(aq_ctx)	(aq_ctx)->aq_chainlen = 0
# define RHTINCRCHAIN(aq_ctx)	do {	\
		if (++(aq_ctx)->aq_chainlen > (aq_ctx)->aq_chainmax)	\
			(aq_ctx)->aq_chainmax = (aq_ctx)->aq_chainlen;\
	} while (0)
# define RHTCHAINLEN(aq_ctx, hd, aq_node) do {	\
		aq_node_P h;		\
		(aq_ctx)->aq_chainlen = 0;	\
		h = aq_node;			\
		while (h != LIST_END(hd)) {				\
			++(aq_ctx)->aq_chainlen;	\
			h = LIST_NEXT(h, ecn_htl);	\
		}				\
		if ((aq_ctx)->aq_chainlen > (aq_ctx)->aq_chainmax)	\
			(aq_ctx)->aq_chainmax = (aq_ctx)->aq_chainlen;	\
	} while (0)

#else /* SM_HT_PERF */
# define aq_rhtincradd(aq_ctx)
# define aq_rhtincrlu(aq_ctx)
# define aq_rhtincrdel(aq_ctx)
# define RHTINCRACOL(aq_ctx, hd)
# define aq_rhtincrlcol(aq_ctx)
# define RHTINCRDCOL(aq_ctx, aq_node)
# define aq_rhtincrgrows(aq_ctx)
# define aq_rhtinitchain(aq_ctx)
# define RHTINCRCHAIN(aq_ctx)
# define RHTCHAINLEN(aq_ctx, hd, aq_node)
#endif /* SM_HT_PERF */

sm_ret_T aq_rht_init(aq_ctx_P _aq_ctx, uint _htsize);
sm_ret_T aq_tht_init(aq_ctx_P _aq_ctx, uint _htsize);
void aq_rht_destroy(aq_ctx_P _aq_ctx);
void aq_tht_destroy(aq_ctx_P _aq_ctx);

#endif /* AQ_ADB_H */
