/*
 * Copyright (c) 2010 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: readfile.c,v 1.6 2011/03/18 18:25:54 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/stat.h"
#include "sm/fcntl.h"
#include "sm/io.h"
#include "sm/misc.h"

/*
**  READFILE -- read a file (entirely)
**
**	Parameters:
**		fn -- file name
**		pcontent -- (pointer to) file content (output)
**		plen -- (pointer to) length of file content (output)
**
**	Returns:
**		usual error code
*/

sm_ret_T
readfile(const char *fn, uchar **pcontent, size_t *plen)
{
	sm_ret_T ret;
	int fd, e;
	char *content;
	size_t len;
	ssize_t rlen;
	struct stat s;

	SM_REQUIRE(fn != NULL);
	SM_REQUIRE(pcontent != NULL);
	SM_REQUIRE(plen != NULL);
	fd = INVALID_FD;
	content = NULL;
	ret = SM_SUCCESS;
	e = stat(fn, &s);
	if (e != 0)
		goto error;

	len = s.st_size;
	content = sm_malloc(len);
	fd = open(fn, O_RDONLY, 0);
	if (fd < 0)
		goto error;
	rlen = read(fd, content, len);
	if (rlen == -1)
		goto error;
	if (rlen != len)
		goto error;
	close(fd);
	*pcontent = (uchar *)content;
	*plen = len;
	return 0;

  error:
	if (SM_SUCCESS == ret)
		ret = sm_err_perm(errno);
	if (fd >= 0)
		close(fd);
	if (content != NULL)
		sm_free(content);
	*pcontent = NULL;
	*plen = 0;
	return ret;
}
