/*
 * Copyright (c) 2010 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sm_inet_ip2arpa.c,v 1.2 2013/09/22 20:23:10 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/net.h"
#include "sm/str.h"
#include "sm/io.h"


/*
**  SM_INET_IP2ARPASTR -- convert an IP address struct to .ARPA format
**
**	Parameters:
**		ipaddr -- IP address struct
**		dst -- append IP6 ARPA ASCII representation (output)
**
**	Returns:
**		usual sm_error
*/

sm_ret_T
sm_inet_ip2arpastr(sm_ipaddr_P ipaddr, sm_str_P dst)
{
	SM_REQUIRE(ipaddr != NULL);

	switch (ipaddr->ipa_family) {
	  case AF_INET:
		return sm_inet_ipv42arpastr(ipaddr->ipa_addr.ipv4_addr, dst);
		break;

#if MTA_NETINET6
	  case AF_INET6:
		return sm_inet_ipv62arpastr(&ipaddr->ipa_addr.ipv6_addr, dst);
		break;
#endif
	  default:
		return sm_error_perm(SM_EM_NONE, EPROTONOSUPPORT);
	}

	/* NOTREACHED */
	return sm_error_perm(SM_EM_NONE, EINVAL);
}
